/* $Id: monitor_main.c,v 1.8 2013/08/29 23:50:09 jrothwei Exp jrothwei $ */
/* Joseph Rothweiler, Sensicomm LLC. Branch from usbio.c on 19Aug2013.
** Standalone program to open the USB device and operate the MHZ100Q2
** 130827 - Changing to new register values.
**
** Endpoints used are:
** 0x83 - Input main data channel.
** 0x04 - Output control channel for command characters.
** 0x85 - Input status interrupt channel.
**
** Command characters for EP 5 are:
** 0-9,a-f - hex bytes
** : - Clear accum
** @ - Copy accum to addr
** = - Copy accum to data
**
** Addresses are:
** 0 - Read address
** 1 - Read length
** 2 - Decimation
** 3 - right shift.
** 4 - dac8step
** 6 - posttrigger
** 8 - Clear selected control bits.
** 9 - Set selected control bits.
** a - strobe lines
**
** Control bits for 8 and 9 are:
** 0 trigger_enable
** 1 trigger_on_rising
** 2 trigger_extern
** 3 channel select (for read)
** 4 pip (old debug thing)

** Strobe bits for address 10 (0xa) are:
** chanx_hilo_reset    0
** ep5_trigger_s48     1
** trigger_manual      2
** chanx_cic_write_set 3 (capture start)
*/
char *m1=":5@:=:2@:14=:3@:e=:4@:83126e=:6@:7d0=";
char *m2=":5@:20=";
char *m3=":5@:8=";
char *m4=":0@:=:1@:100=";

char *cmdep5 = ":a@:2="; /* reg10 = 0x2 : bit 1 = ep5_trigger */
char *cmd6clear = ":6@:2=";

// This works: sets posttrigger to 7d0 and holds.
char *cmd5c6v=":8@:1f=:6@:7d0="; // r8=0x1f (clear all flags) r6=2000 (0x7d0).
char *cmdwrites=":a@:8=";    // Set cic_write.
char *cmdmtrig=":a@:4=:=";      // Manual trigger.
char *cmdextrig=":8@:1= :a@:8= :9@:4=";  // external trigger.
                 //R8=1 clear trig en, Ra=8 capture start. R9=4 external trigger.
char *cmdgetdata=":0@:=:1@:100="; // Reg0=0; Reg1=0x100 - read 256 bytes from address 0.
// char *cmd5c6v=":5@:=:6@:7d0=:5@:8="; // r5=0 (clear all flags) r6=2000 (0x7d0). r5=8 (manual trig).
char *cmdgrab=":5@:20=";    // Set cic_write.

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>        /* For mode constants */
#include <fcntl.h>           /* For O_* constants */
#include <time.h>
#include <mqueue.h>
#include <ctype.h>

#include <libusb-1.0/libusb.h>

#include "usb_lib.h"

#define ArraySize(x) (sizeof(x)/sizeof(*x))

// sh_ptr->vals[*] follow the VHDL code's internal register numbering.

enum vhdl_regs {
  VREG_READADDR=0 , VREG_READLEFT=1, VREG_CIC     =2, VREG_RTSHIFT=3,
  VREG_DDS     =4,  VREG_FLAGS   =5, VREG_POSTTRIG=6
};

int do_reattach = 0;

int exitflag = 0;
void mysigint(int signum) {
	exitflag = 1;
}
void printit(char *buf, int len) {
	int k;
	for(k=0;k<len;k++) {
		if(k%16==0) fprintf(stderr,"%4x:",k);
		fprintf(stderr," %2.2x",buf[k]);
		if(k%16==15) fprintf(stderr,"\n");
	}
	if(k%16!=0) fprintf(stderr,"\n");
}
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
// Registers:
// 0 read address
// 1 read count
// 2 cic factor - 1
// 3 cic output rightshift
// 4 DDS step.
// 5 strobes:
//   0: hilo_reset     (level; should be edge).
//   2: trigger enable (level; should be edge).
//   4: 1:rising edge, 0: falling edge.
//   8: Manual trigger (level; should be edge).
//  10: EP5 (USB interrupt channel) trigger
//  20: Set cic_write (start capturing; ends on post-trigger counter going to zero).
// 6 lag (number of samples captured after trigger)
// Initial setup:
int8_t setupcommand[] =
// 12345678
  ":@=1@00="  // 8 Reg0 = 0; Reg1 = 0x100
  ":2@:13="   // 7 Reg2 = 19 (0x13) cic=20
  ":3@:e="    // 6 Reg3 = 14 (0xe) shift=14
  ":4@f4240=" // 9 Reg4=1000000 (0xf4240) dds=
  ":6@7d0="   // 7 Reg6=2000 (0x7d0) lag=
  ":5@29="    // 6 Reg5=0x29 - Reset; load post-trigger counter. This is a VHDL bug workaround.
  ":5@20="    // 6 Reg5=0x20 - Working mode.
;
int8_t setupbasic[] =
// 12345678
  ":@=1@00="  // 8 Reg0 = 0; Reg1 = 0x100   - Start reads at address 0, length 256 (0x100).
  ":2@:="   // 7 Reg2 = 0 cic=1
  ":3@:="    // 6 Reg3 = 0 shift=0
  ":4@f4240=" // 9 Reg4=1000000 (0xf4240) dds=
  ":6@7d0="   // 7 Reg6=2000 (0x7d0) lag=
  ":5@29="    // 6 Reg5=0x29 - Reset; load post-trigger counter. This is a VHDL bug workaround.
  ":5@20=:="    // 6 Reg5=0x20,0 - Working mode.
;
int8_t outdata[64];
int outdatalen = 49;
/*--.----|----.----|----.----|----.----|----.----|----.----|----.----|----.---*/
uint64_t llpackvals(int8_t *data, int len, int beg, int end) {
	uint64_t rtn;
	int k;
	rtn = 0;
	if(end>=len) return rtn;
	for(k=beg;k<=end;k++) rtn = (rtn<<8) | ((uint8_t *)data)[k];
	return rtn;
}
uint32_t packvals(int8_t *data, int len, int beg, int end) {
	return (uint32_t)llpackvals(data, len, beg, end);
}
void ep5_display(int8_t *data, int len) {
	int k;
	if(0) {
		fprintf(stdout,"EP5 %d:",len);
		for(k=0;k<len;k++) fprintf(stdout," %2.2x",0xff&data[k]);
		fprintf(stdout,"\n");
	}
	// Breakouts, updated for rev 1.18.
	uint32_t r1      =          packvals(data,len, 0, 3);
	uint32_t r2      =          packvals(data,len, 4, 7);
	uint32_t zip1    =          packvals(data,len, 8,11);
	uint32_t posttrg =          packvals(data,len,12,13)>>4;
	uint32_t cicwrit =          packvals(data,len,13,13)&0xf; // Really 1 bit.
	uint32_t capcnt  =          packvals(data,len,14,14)>>4;
	uint32_t wr_addr =          packvals(data,len,14,15)&0xfff;
	uint32_t cmddata =          packvals(data,len,16,19);
	int32_t  adc0max = (int32_t)packvals(data,len,20,20);
	int32_t  adc0min = (int32_t)packvals(data,len,21,21);
	uint64_t cic0min =          llpackvals(data,len,22,26);
	uint64_t cic0max =          llpackvals(data,len,27,31);
	fprintf(stdout," post %3x cic %x cap %x wr %3x cmd %8x adc %2x %2x cic %10lx %10lx",
		posttrg,cicwrit,capcnt,wr_addr,cmddata,adc0min,adc0max,cic0min,cic0max);
	if(r1!=0x01234567) fprintf(stdout," r1 0x%x",r1);
	if(r2!=0x89abcdef) fprintf(stdout," r2 0x%x",r2);
	if(zip1!=0) fprintf(stdout," zip1 0x%x",zip1);
	fprintf(stdout,"\n");
}
static void sendcmd(libusb_device_handle *handle,int endp,char *cmd) {
	int actlen;
	int timeout_ms = 100;
	checked_bulk_transfer (handle,endp,
		cmd,strlen(cmd),
		&actlen,timeout_ms,0x20);
}
	
int main(int argc, char *argv[]) {
	int rtn;
	libusb_context *context;
	int ofile = -1;
	int k;
	int intf_num = 0;
	int intf_alt = 0;
	uint8_t ep;

	/* Change stdin to nonblocking mode. */

	int flags;
	flags = fcntl(0, F_GETFL); 
	flags |= O_NONBLOCK;	
	fcntl(0, F_SETFL, flags);

	for(k=1;k<argc;k++) {
		if(strncmp("-o",argv[k],2)==0) {
			ofile=open(argv[k]+2,O_WRONLY|O_CREAT|O_EXCL,0644);
			if(ofile < 0) {
				fprintf(stderr,"Trouble creating output file <%s>\n",argv[k]+2);
				exit(2);
			}
		} else if(strncmp("-e",argv[k],2)==0) {
			ep = strtoul(argv[k]+2,0,0);
		} else {
			fprintf(stderr,"Bad argument <%s>\n",argv[k]);
			exit(2);
		};
	}
	if(ofile<0) {
		fprintf(stderr,"No output file\n");
		sleep(1);
	}
	// fprintf(stdout,"A\n"); fflush(stdout);

	signal(SIGINT,mysigint);

	/******************************
	** Init libusb-1.0.          */

	rtn = libusb_init (&context);
	if(rtn != 0) {
		fprintf(stderr,"%s: init trouble %d\n",argv[0],rtn);
		exit(2);
	}
	libusb_set_debug(context,3); // Info to stdout, warn/error to stderr.
	// fprintf(stdout,"C\n"); fflush(stdout);

	/******************************
	** Find a useful device. */

	libusb_device *found = NULL;
	found = find_usb_device(context,0x00fe,0x2100,1) ;
	if (!found) {
		fprintf(stderr,"Device not found\n");
		exit(2);
	} else {
		fprintf(stderr,"Found it\n");
	}
	// fprintf(stdout,"D\n"); fflush(stdout);

	libusb_device_handle *handle = NULL;
	// fprintf(stdout,"E\n"); fflush(stdout);
	if (found) {
		int err;

		fprintf(stderr,"Opening\n");
		err = libusb_open(found, &handle);
		if (err) {
			fprintf(stderr,"Trouble opening device %d %s\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F1\n"); fflush(stdout);
		err = libusb_kernel_driver_active(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"No kernel driver active.\n");
		} else if(err == 1) {
			fprintf(stderr,"Releasing kernel driver.\n");
			err = libusb_detach_kernel_driver(handle,intf_num);
			if(err == 0) {
				fprintf(stderr,"kernel driver successfully detached\n");
				do_reattach = 1;
			} else {
				fprintf(stderr,"Error code %d %s detaching driver\n",err,errorname(err));
				exit(2);
			}
		} else {
			fprintf(stderr,"error %d %s from libusb_kernel_driver_active\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F2\n"); fflush(stdout);

		fprintf(stderr,"Claiming\n");
		err = libusb_claim_interface(handle,intf_num);
		if (err) {
			fprintf(stderr,"Trouble claim_interface %d %s\n",err,errorname(err));
			exit(2);
		}
	// fprintf(stdout,"F3\n"); fflush(stdout);
		usb_dumpinfo(found);
		fprintf(stderr,"Alt %d\n",intf_alt);
		err = libusb_set_interface_alt_setting (handle,intf_num,intf_alt);
		if (err) {
			fprintf(stderr,"Trouble alt_setting %d %s\n",err,errorname(err));
			exit(2);
		}
		fprintf(stderr,"Xfer 0x%2.2x\n",ep);
		// continuous_transfer(handle,ofile,ep);
		int loopcount;
		for(loopcount=0;!exitflag;loopcount++) {
			int8_t indata[512];
			// This is just a test.
			// int8_t outdata[] = ":123456=@            ";
			// This partly works: Starts continuous send.
			//int8_t outdata[] = ": @ = :1 @ :100 = :  "; // 0<=0 1<=0x100 (Read 256 from 0)
			// int8_t outdata[] = ":@=:1@:100=:7@:=         "; // Not working.
			int actlen;
			int timeout_ms = 100;
			// Main data channel.
			checked_bulk_transfer (handle,0x83,(char *)indata,512,&actlen,timeout_ms,1);
			if(actlen>0) {
				fprintf(stdout,"Data in %d\n",actlen);
			}
			// Int channel.
			checked_bulk_transfer (handle,0x85,(char *)indata,512,&actlen,timeout_ms,0);
			if(actlen>0) {
				// fprintf(stdout,"Chan 5 in %d\n",actlen);
				ep5_display(indata,actlen);
			}
			// Out control channel.
			// if(msgflag) {
			// 	fprintf(stdout,"REQ\n");
			// 	checked_bulk_transfer (handle,0x04,outdata,outdatalen,&actlen,timeout_ms,0x23);
			// 	msgflag = 0;
			// }
			if( (loopcount%10) == 0) {
				fprintf(stdout,"****** Loop %d\n",loopcount);
				sendcmd(handle,0x4,cmdep5);
			}
			if(1) {
				if( loopcount == 15) sendcmd(handle,0x4,cmd5c6v);
				if( loopcount == 35) sendcmd(handle,0x4,cmdwrites); // wr addr changes.
				if( loopcount == 55) sendcmd(handle,0x4,cmdmtrig);  // wr stops.
				if( loopcount == 75) sendcmd(handle,0x4,cmdgetdata);
				if( loopcount == 95) sendcmd(handle,0x4,cmdextrig);
			}
			if(0) {
				// Verify sequence for Octave code.
				if( loopcount == 15) sendcmd(handle,0x4,m1);
				if( loopcount == 35) sendcmd(handle,0x4,m2);
				if( loopcount == 55) sendcmd(handle,0x4,m3);
				if( loopcount == 75) sendcmd(handle,0x4,m4);
				if( loopcount == 100) break;
			}
			// if( loopcount == 25) sendcmd(handle,0x4,cmdgrab);
			// Dummy for now.
			if( (loopcount) == 10000) {
				checked_bulk_transfer (handle,0x04,
					(char *)setupcommand,ArraySize(setupcommand)-1,
					// setupbasic,ArraySize(setupbasic)-1,
					&actlen,timeout_ms,0x23);
			}
			if( (loopcount) == 40000) {
				checked_bulk_transfer (handle,0x04,
					cmd6clear,ArraySize(cmd6clear)-1,
					&actlen,timeout_ms,0x23);
			}
		}
	} else {
		fprintf(stderr,"Not found.\n");
	}
	if(do_reattach) {
		int err;
		err = libusb_attach_kernel_driver(handle,intf_num);
		if(err == 0) {
			fprintf(stderr,"kernel driver reattached\n");
		} else {
			fprintf(stderr,"Error %d %s on kernel driver reattach\n",err,errorname(err));
		}
	} else {
		fprintf(stderr,"No driver to reattach\n");
	}

	/******************************
	** Done. */
	if(handle != NULL) libusb_close(handle);
	libusb_exit (context);
	return 0;
}
